// written by Zheng Gengbin and Shi Lixia
// to process ALLOCATE & DEALLOCATE

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include "compilerApi.h"
#include "anls_att.h"

extern  int do_level ;

SgSymbol *get_symbol(SgStatement *s, SgType *type)
{  
  SgSymbol * symbol=NULL;

  switch(s->variant())
   {
     case ALLOCATE_STMT:
        symbol=s->getSymbolInCurrentScope(mixstring,allocNameOfType(type),NULL);
        break;
     case DEALLOCATE_STMT:
        symbol=s->getSymbolInCurrentScope(mixstring,freeNameOfType(type),NULL);
        break;
     case ASSIGN_STAT:
        symbol=s->getSymbolInCurrentScope(mixstring,stackNameOfType(type),NULL);
        break;
   }

  return symbol;
}

///////// real,allocatable::a(:)-------->integer::pcrc_index_a   

void analyse_allocatable(SgStatement * s)
{
  if(s->expr(2)!=NULL)
     for(SgExpression *exp=s->expr(2); exp; exp=exp->rhs())
        {   //assume VAR_DECL has only  ALLOCATABLE_OP in expr(2)
         switch(exp->lhs()->variant())
           {
             case ALLOCATABLE_OP:
                { 
/*
		  SgExpression *  expr1=NULL;
                  for(SgExpression *expression=s->expr(0);expression;expression=expression->rhs())
                   {
                    SgExpression * var_ref=new SgExpression(VAR_REF,NULL,NULL,NULL);
                    SgSymbol     * symbol=s->getSymbolInCurrentScope(mixstring,"index",expression->lhs()->symbol()->identifier());
                    var_ref->setSymbol(symbol);
                    expr1=new SgExpression(EXPR_LIST,var_ref,expr1,NULL);
                        
                   } 
                  SgType       * type=SgTypeInt();
                  SgExpression * expr2=new SgExpression(TYPE_OP,NULL,NULL,NULL,type);
                  SgStatement  * declStmt=new SgStatement(VAR_DECL,NULL,NULL,expr1,expr2,NULL);
                  s->insertStmtAfter(*declStmt,*s->controlParent());
*/
                  s->deleteStmt();   
                  break;
               }
	    case POINTER_OP:
                  s->deleteStmt();   
                  break;
          }
       }           
   
}

void analyse_allocate(SgStatement * s)
{
   for(SgExpression * expr=s->expr(0); expr; expr=expr->rhs())  
   {
	SgExpression *exp = expr->lhs();
	SgSymbol *symb = exp->symbol();
	symb->deleteAllAttribute(Allocate_Def);
	symb->addAttribute(Allocate_Def, exp, sizeof(exp));
	if (symb->numberOfAttributes(Dist_Def) > 0) {
	  Distribute_def *ddef = (Distribute_def *)symb->attributeValue(0, Dist_Def);
        }
   }
}

//------allocate(a(global_size))------>pcrc_index_a=pcrc_real_alloc(local_size)
void trans_allocate(SgStatement * s)
{
   SgExpression *expr = s->expr(0)->lhs();
   SgSymbol *symb = expr->symbol();
     // record size 
   symb->deleteAllAttribute(Allocate_Def);
   symb->addAttribute(Allocate_Def, expr, sizeof(expr));
     // only deal with scalar pointer
     // array allocation is performed by re-align & re-distribute afterwards
   if (is_pointer(symb) && isSgArrayType(symb->type()) == NULL)
   {
	SgType *type = symb->type();
	SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", symb->identifier());
	SgStatement *new_data = make_new_array_data(s, dad, type, -1, 0, NULL);
	s->insertStmtBefore(*new_data, *s->controlParent());

        SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", symb->identifier(), 0);
   	SgSymbol * mHandle = s->getSymbolInCurrentScope(mixstring, "mh", symb->identifier(), 0);

          // pcrc_stack_alloc
   	SgStatement *allocstmt=make_heap_alloc(s,symb->type(),base,dad,mHandle);
   	s->insertStmtBefore(*allocstmt, *s->controlParent());

    	  // reset_array_base
   	SgStatement *reset_stmt = make_reset_array_base(s,dad,type,base, mHandle);
   	s->insertStmtBefore(*reset_stmt, *s->controlParent());
   }
   s->deleteStmt();
}

#if 0
void trans_allocate(SgStatement * s)
{
   for(SgExpression * exp=s->expr(0); exp; exp=exp->rhs())  
     {
       SgSymbol * lhs_symbol=s->getSymbolInCurrentScope(mixstring,"index",exp->lhs()->symbol()->identifier());
       SgSymbol * rhs_symbol=NULL; 
       rhs_symbol=get_symbol(s,exp->lhs()->symbol()->type()->baseType());
      SgType       * type=SgTypeInt();
      SgExpression * expr1=new SgExpression(VAR_REF,NULL,NULL,lhs_symbol,type);
      SgSymbol     * symbol_1=s->getSymbolInCurrentScope(mixstring,"size",NULL); 
      SgSymbol     * symbol_2=s->getSymbolInCurrentScope(mixstring,"dad",exp->lhs()->symbol()->identifier());
      SgExpression * expression=new SgExpression(VAR_REF,NULL,NULL,symbol_2,type);
      SgExpression * exprlist=NULL;
      exprlist=new SgExpression(EXPR_LIST,expression,NULL,NULL);
      exprlist=new SgExpression(FUNC_CALL,exprlist,NULL,symbol_1);
      exprlist=new SgExpression(EXPR_LIST,exprlist,NULL,NULL);
      SgExpression * expr2=new SgExpression(FUNC_CALL,exprlist,NULL,rhs_symbol);
      SgStatement  * assignStmt=new SgStatement(ASSIGN_STAT,NULL,NULL,expr1,expr2,NULL);
      s->insertStmtAfter(*assignStmt,*s->controlParent());
       
     }
   s->deleteStmt();
}
#endif


///    deallocate(a)------->call pcrc_real_free(pcrc_index_a)
  
void trans_deallocate(SgStatement * s)
{
   for(SgExpression * expr=s->expr(0); expr; expr = expr->rhs())  
   {
      SgExpression *exp = expr->lhs();
      SgSymbol *symb = exp->symbol();
      SgSymbol *dad=s->getSymbolInCurrentScope(mixstring,"dad",symb->identifier());

	// delete dad
      SgStatement *deleteStmt = make_delete_array(s, dad);
      s->insertStmtAfter(*deleteStmt, *s->controlParent());

        // delete the allocated memory of old dad
      SgStatement *freeStmt = make_array_free(s, dad);
      s->insertStmtAfter(*freeStmt, *s->controlParent());
   }
   s->deleteStmt();
}

// assign -1 to dad as nullify
void trans_nullify(SgStatement * s)
{
   for(SgExpression * expr=s->expr(0); expr; expr = expr->rhs())  
   {
      SgExpression *exp = expr->lhs();
      SgSymbol *symb = exp->symbol();
      SgSymbol *dad=s->getSymbolInCurrentScope(mixstring,"dad",symb->identifier());

      SgExpression *dadExp = new SgVarRefExp(dad);
      SgExpression *val = new SgValueExp(-1);
      SgAssignStmt *stmt = new SgAssignStmt(*dadExp, *val);
      s->insertStmtAfter(*stmt, *s->controlParent());
   }
   s->deleteStmt();
}

//--------- replace variable that has allocatable attribute

void transform_3(SgStatement * s)
{ 
     switch(s->variant())
     {
      case GLOBAL:
       {
        int org_numberOfChildren=s->numberOfChildrenList1();
        SgStatement * * children = new SgStatement * [org_numberOfChildren];
      
        int i;
  
        for (i=0; i<org_numberOfChildren; i++)
          children[i]=s->childList1(i);

        for (i=0; i<org_numberOfChildren; i++)
          transform_3(children[i]);
      
        delete [] children;
        break;
       };
 
     case PROG_HEDR:
       {
         in_main_program = on;
         int org_numberOfChildren=s->numberOfChildrenList1();
         SgStatement * * children = new SgStatement * [org_numberOfChildren];

         int i;

         for (i=0; i<org_numberOfChildren; i++)
             children[i]=s->childList1(i);

         for (i=0; i<org_numberOfChildren; i++)
              transform_3(children[i]);
      
         delete [] children;
      
          in_main_program=off;

         break;
       };
      case PROC_HEDR:
        {
          SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
          int org_numberOfChildren=s->numberOfChildrenList1();
          SgStatement * * children = new SgStatement * [org_numberOfChildren];

          int i;

          for (i=0; i<org_numberOfChildren; i++)
                children[i]=s->childList1(i);

          in_procedure=on;

          for (i=0; i<org_numberOfChildren; i++)
                 transform_3(children[i]);

          delete [] children;
      
          in_procedure = off;
          break;
          };
       case FOR_NODE:
         {
           do_level++;

           int org_numberOfChildren=s->numberOfChildrenList1();

            SgStatement * * children = new SgStatement * [org_numberOfChildren];

           int i;

           for (i=0; i<org_numberOfChildren; i++)
               children[i]=s->childList1(i);

             for (i=0; i<org_numberOfChildren; i++)
               transform_3(children[i]);
      
            delete [] children;
      
           do_level--;

            break;
          };
      case IF_NODE:
          {
            int org_numberOfChildren1=s->numberOfChildrenList1();

            SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];
 
            int org_numberOfChildren2=s->numberOfChildrenList2();

            SgStatement * * children2 = new SgStatement * [org_numberOfChildren2];

            int i;

            for (i=0; i<org_numberOfChildren1; i++)
                 children1[i]=s->childList1(i);

            for (i=0; i<org_numberOfChildren1; i++)
                 transform_3(children1[i]);
      
            delete [] children1;

            for (i=0; i<org_numberOfChildren2; i++)
                children2[i]=s->childList2(i);

            for (i=0; i<org_numberOfChildren2; i++)
                 transform_3(children2[i]);
      
            delete [] children2;

            break;
          } ;
       case ASSIGN_STAT:
           {  
             SgSymbol     * symbol=NULL;
             SgType       * type=SgTypeInt();
             char         * ident=NULL;
             SgExpression * expression=NULL;

//printf("s->expr(0)->symbol():%s\n",s->expr(0)->symbol()->identifier());

               //lhs of assign statement
             if(s->expr(0)->symbol()->attributes() & ALLOCATABLE_BIT)
             {
                symbol=get_symbol(s, s->expr(0)->symbol()->type()->baseType());
                ident=s->expr(0)->symbol()->identifier();
                s->expr(0)->setSymbol(symbol); 
                symbol=s->getSymbolInCurrentScope(mixstring,"index",ident);
                expression=new SgExpression(VAR_REF,NULL,NULL,symbol,type);
                expression=new SgExpression(ADD_OP,expression,s->expr(0)->lhs()->lhs(),NULL,type);
                s->expr(0)->lhs()->setLhs(expression);
              } 

                 // rhs of assign statement
              for(SgExpression * exp=s->expr(1); exp; exp=exp->lhs())
                { 
//printf("exp->symbol:%d\n",exp->variant());
//if (exp->variant() == VAR_REF) printf("exp->symbol:%s\n",exp->symbol()->identifier());
                  if(exp->variant()==ARRAY_REF)
                   {
                     if(exp->symbol()->attributes() & ALLOCATABLE_BIT)    
                      {
                       symbol=get_symbol(s, exp->symbol()->type()->baseType());
                       ident=exp->symbol()->identifier();
                        exp->setSymbol(symbol); 
                       symbol=s->getSymbolInCurrentScope(mixstring,"index",ident);
                       expression=new SgExpression(VAR_REF,NULL,NULL,symbol,type);
                       expression=new SgExpression(ADD_OP,expression,exp->lhs()->lhs(),NULL,type);
                       exp->lhs()->setLhs(expression);
                      }
                   }
                  else if(exp->rhs()!=NULL)
                        if(exp->rhs()->variant()==ARRAY_REF)
                       {
                         if(exp->rhs()->symbol()->attributes() & ALLOCATABLE_BIT)    
                         {
                          symbol=get_symbol(s, exp->rhs()->symbol()->type()->baseType());
                          ident=exp->rhs()->symbol()->identifier();
                           exp->rhs()->setSymbol(symbol); 
                          symbol=s->getSymbolInCurrentScope(mixstring,"index",ident);
                           expression=new SgExpression(VAR_REF,NULL,NULL,symbol,type);
                          expression=new SgExpression(ADD_OP,expression,exp->rhs()->lhs()->lhs(),NULL,type);
                          exp->rhs()->lhs()->setLhs(expression);
                         }
                       }//end of else
                 }  //end of for
              break;
           };     
      }
}




